@@ -162,20 +162,6 @@ int btrfs_csum_data(struct btrfs_fs_info *fs_info, u16 csum_type, const u8 *data
 	return -1;
 }
 
-int btrfs_format_csum(u16 csum_type, const u8 *data, char *output)
-{
-	int i;
-	const int csum_size = btrfs_csum_type_size(csum_type);
-
-	sprintf(output, "0x");
-	for (i = 0; i < csum_size; i++) {
-		output += 2;
-		sprintf(output, "%02x", data[i]);
-	}
-
-	return csum_size;
-}
-
 static int __csum_tree_block_size(struct extent_buffer *buf, u16 csum_size,
 				  int verify, int silent, u16 csum_type)
 {
@@ -189,9 +175,8 @@ static int __csum_tree_block_size(struct extent_buffer *buf, u16 csum_size,
 	if (verify) {
 		if (memcmp_extent_buffer(buf, result, 0, csum_size)) {
 			if (!silent) {
-				/* "0x" plus 2 hex chars for each byte plus nul */
-				char found[2 + BTRFS_CSUM_SIZE * 2 + 1];
-				char wanted[2 + BTRFS_CSUM_SIZE * 2 + 1];
+				char found[BTRFS_CSUM_STRING_LEN];
+				char wanted[BTRFS_CSUM_STRING_LEN];
 
 				btrfs_format_csum(csum_type, result, found);
 				btrfs_format_csum(csum_type, (u8 *)buf->data, wanted);
@@ -720,7 +705,10 @@ out:
 		return ERR_PTR(-EIO);
 	}
 insert:
-	root->ref_cows = 1;
+	if (root->root_key.objectid != BTRFS_TREE_LOG_OBJECTID)
+		root->track_dirty = 1;
+	if (is_fstree(root->root_key.objectid))
+		root->ref_cows = 1;
 	return root;
 }
 
